"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.resolveModuleOptions = exports.createRawModuleRuleUses = void 0;
const assert_1 = __importDefault(require("assert"));
const path_1 = __importDefault(require("path"));
const utils_1 = require("../utils");
const devtool_1 = require("./devtool");
const toBuffer = (bufLike) => {
    if (Buffer.isBuffer(bufLike)) {
        return bufLike;
    }
    else if (typeof bufLike === "string") {
        return Buffer.from(bufLike);
    }
    throw new Error("Buffer or string expected");
};
function composeJsUse(uses, options) {
    if (!uses.length) {
        return null;
    }
    async function loader(data) {
        var _a;
        let cacheable = data.cacheable;
        let content = data.content;
        let sourceMap = (_a = data.sourceMap) === null || _a === void 0 ? void 0 : _a.toString("utf-8");
        let additionalData = data.additionalData
            ? JSON.parse(data.additionalData.toString("utf-8"))
            : undefined;
        // Loader is executed from right to left
        for (const use of uses) {
            (0, assert_1.default)("loader" in use);
            let loaderResult;
            const p = new Promise((resolve, reject) => {
                let isDone = false;
                // Whether a `callback` or `async` is called
                let isSync = true;
                let isError = false; // internal error
                let reportedError = false;
                function callback(err, content, sourceMap, additionalData) {
                    if (isDone) {
                        if (reportedError)
                            return; // ignore
                        err = new Error("callback(): The callback was already called.");
                    }
                    isSync = false;
                    isDone = true;
                    if (err) {
                        isError = true;
                        reject(err);
                        return;
                    }
                    resolve({
                        cacheable,
                        content,
                        sourceMap,
                        additionalData
                    });
                }
                const loaderContext = {
                    sourceMap: (0, devtool_1.isUseSourceMap)(options.devtool),
                    resourcePath: data.resourcePath,
                    resource: data.resource,
                    // Return an empty string if there is no query or fragment
                    resourceQuery: data.resourceQuery || "",
                    resourceFragment: data.resourceFragment || "",
                    getOptions() {
                        return use.options;
                    },
                    cacheable(value) {
                        cacheable = value;
                    },
                    async() {
                        if (isDone) {
                            if (reportedError)
                                return; // ignore
                            reject(new Error("async(): The callback was already called."));
                        }
                        isSync = false;
                        return callback;
                    },
                    callback,
                    rootContext: options.context,
                    context: path_1.default.dirname(data.resourcePath)
                };
                /**
                 * support loader as string
                 */
                let loader;
                if (typeof use.loader === "string") {
                    try {
                        let loaderPath = require.resolve(use.loader, {
                            paths: [options.context]
                        });
                        loader = require(loaderPath);
                    }
                    catch (err) {
                        reject(err);
                        return;
                    }
                }
                else {
                    loader = use.loader;
                }
                let result = undefined;
                try {
                    result = loader.apply(loaderContext, [
                        loader.raw ? Buffer.from(content) : content.toString("utf-8"),
                        sourceMap,
                        additionalData
                    ]);
                    if (isSync) {
                        isDone = true;
                        if (result === undefined) {
                            resolve({
                                content,
                                sourceMap,
                                additionalData,
                                cacheable
                            });
                            return;
                        }
                        if ((0, utils_1.isPromiseLike)(result)) {
                            return result.then(function (result) {
                                resolve({
                                    content: result,
                                    sourceMap,
                                    additionalData,
                                    cacheable
                                });
                            }, reject);
                        }
                        return resolve({
                            content: result,
                            sourceMap,
                            additionalData,
                            cacheable
                        });
                    }
                }
                catch (err) {
                    if (isError) {
                        reject(err);
                        return;
                    }
                    if (isDone) {
                        // loader is already "done", so we cannot use the callback function
                        // for better debugging we print the error on the console
                        if (typeof err === "object" && err.stack)
                            console.error(err.stack);
                        else
                            console.error(err);
                        reject(err);
                        return;
                    }
                    isDone = true;
                    reportedError = true;
                    reject(err);
                }
            });
            if ((loaderResult = await p)) {
                additionalData =
                    (typeof loaderResult.additionalData === "string"
                        ? JSON.parse(loaderResult.additionalData)
                        : loaderResult.additionalData) || additionalData;
                content = loaderResult.content || content;
                sourceMap = loaderResult.sourceMap || sourceMap;
                cacheable = loaderResult.cacheable || cacheable;
            }
        }
        return {
            cacheable: cacheable,
            content: toBuffer(content),
            sourceMap: sourceMap
                ? toBuffer(typeof sourceMap === "string"
                    ? sourceMap
                    : JSON.stringify(sourceMap))
                : undefined,
            additionalData: additionalData
                ? toBuffer(JSON.stringify(additionalData))
                : undefined
        };
    }
    loader.displayName = `NodeLoaderAdapter(${uses
        .map(item => {
        (0, assert_1.default)("loader" in item);
        let loader;
        if (typeof item.loader === "string") {
            try {
                const path = require.resolve(item.loader, {
                    paths: [options.context]
                });
                loader = require(path);
            }
            catch (e) {
                loader = null;
            }
        }
        else {
            loader = item.loader;
        }
        return (loader === null || loader === void 0 ? void 0 : loader.displayName) || (loader === null || loader === void 0 ? void 0 : loader.name) || "unknown-loader";
    })
        .join(" -> ")})`;
    return {
        loader
    };
}
function createRawModuleRuleUses(uses, options) {
    return createRawModuleRuleUsesImpl([...uses].reverse(), options);
}
exports.createRawModuleRuleUses = createRawModuleRuleUses;
function createRawModuleRuleUsesImpl(uses, options) {
    if (!uses.length) {
        return [];
    }
    const index = uses.findIndex(use => "builtinLoader" in use);
    if (index < 0) {
        return [composeJsUse(uses, options)];
    }
    const before = uses.slice(0, index);
    const after = uses.slice(index + 1);
    return [
        composeJsUse(before, options),
        createNativeUse(uses[index]),
        ...createRawModuleRuleUsesImpl(after, options)
    ].filter((item) => Boolean(item));
}
function createNativeUse(use) {
    var _a;
    (0, assert_1.default)("builtinLoader" in use);
    if (use.builtinLoader === "sass-loader") {
        ((_a = use.options) !== null && _a !== void 0 ? _a : (use.options = {})).__exePath = require.resolve(`@tmp-sass-embedded/${process.platform}-${process.arch}/dart-sass-embedded/dart-sass-embedded${process.platform === "win32" ? ".bat" : ""}`);
    }
    return {
        builtinLoader: use.builtinLoader,
        options: JSON.stringify(use.options)
    };
}
function resolveModuleRuleCondition(condition) {
    if (typeof condition === "string") {
        return {
            type: "string",
            matcher: condition
        };
    }
    if (condition instanceof RegExp) {
        return {
            type: "regexp",
            matcher: condition.source
        };
    }
    throw new Error(`Unsupported condition type ${typeof condition}, value: ${condition}`);
}
function resolveModuleRuleConditions(conditions) {
    return conditions.map(resolveModuleRuleCondition);
}
function resolveModuleOptions(module = {}, options) {
    var _a;
    const rules = ((_a = module.rules) !== null && _a !== void 0 ? _a : []).map(rule => {
        // FIXME: use error handler instead of throwing
        if (rule === null || rule === void 0 ? void 0 : rule.loader) {
            throw new Error("`Rule.loader` is not supported, use `Rule.use` instead");
        }
        if (rule === null || rule === void 0 ? void 0 : rule.uses) {
            throw new Error("`Rule.uses` is deprecated for aligning with webpack, use `Rule.use` instead");
        }
        return {
            ...rule,
            test: (0, utils_1.isNil)(rule.test) ? null : resolveModuleRuleCondition(rule.test),
            include: (0, utils_1.isNil)(rule.include)
                ? null
                : Array.isArray(rule.include)
                    ? resolveModuleRuleConditions(rule.include)
                    : [resolveModuleRuleCondition(rule.include)],
            exclude: (0, utils_1.isNil)(rule.exclude)
                ? null
                : Array.isArray(rule.exclude)
                    ? resolveModuleRuleConditions(rule.exclude)
                    : [resolveModuleRuleCondition(rule.exclude)],
            resource: (0, utils_1.isNil)(rule.resource)
                ? null
                : resolveModuleRuleCondition(rule.resource),
            resourceQuery: (0, utils_1.isNil)(rule.resourceQuery)
                ? null
                : resolveModuleRuleCondition(rule.resourceQuery),
            use: createRawModuleRuleUses(rule.use || [], options)
        };
    });
    return {
        parser: module.parser,
        rules
    };
}
exports.resolveModuleOptions = resolveModuleOptions;
//# sourceMappingURL=module.js.map